from typing import Any, Dict, Optional

from nonebot.exception import ActionFailed
from nonebot.utils import logger_wrapper

log = logger_wrapper("ntchat")


def handle_api_result(result: Optional[Dict[str, Any]]) -> Any:
    """处理 API 请求返回值。

    参数:
        result: API 返回数据

    返回:
        API 调用返回数据

    异常:
        ActionFailed: API 调用失败
    """
    if isinstance(result, dict):
        if result.get("status") == "failed":
            raise ActionFailed(**result)
        return result.get("data")



"""
消息类型枚举，参考：https://www.showdoc.com.cn/579570325733136/3417087407035329
"""

from enum import IntEnum


class EventType(IntEnum):
    """消息类型枚举"""

    MT_DEBUG_LOG = 11024
    """DEBUG消息"""
    MT_RECV_QRCODE_MSG = 11087
    """获取用户登录二维码"""
    MT_USER_LOGIN = 11025
    """登录消息"""
    MT_USER_LOGOUT = 11026
    """注销消息"""
    MT_DATA_FRIENDS_MSG = 11030
    """获取好友列表消息"""
    MT_DATA_CHATROOMS_MSG = 11031
    """获取群聊列表消息"""
    MT_DATA_CHATROOM_MEMBERS_MSG = 11032
    """获取群成员消息"""
    MT_DATA_PUBLICS_MSG = 11033
    """获取公众号消息"""
    MT_SEND_TEXTMSG = 11036
    """发送文本消息"""
    MT_SEND_CHATROOM_ATMSG = 11037
    """发送群@消息"""
    MT_SEND_CARDMSG = 11038
    """发送名片消息"""
    MT_SEND_LINKMSG = 11039
    """发送链接消息"""
    MT_SEND_IMGMSG = 11040
    """发送图片消息"""
    MT_SEND_FILEMSG = 11041
    """发送文件消息"""
    MT_SEND_VIDEOMSG = 11042
    """发送视频消息"""
    MT_SEND_GIFMSG = 11043
    """发送GIF消息"""
    MT_RECV_TEXT_MSG = 11046
    """接收文本消息"""
    MT_RECV_PICTURE_MSG = 11047
    """接收图片消息"""
    MT_RECV_VOICE_MSG = 11048
    """接收视频消息"""
    MT_RECV_FRIEND_MSG = 11049
    """接收申请好友消息"""
    MT_RECV_CARD_MSG = 11050
    """接收名片消息"""
    MT_RECV_VIDEO_MSG = 11051
    """接收视频消息"""
    MT_RECV_EMOJI_MSG = 11052
    """接收表情消息"""
    MT_RECV_LOCATION_MSG = 11053
    """接收位置消息"""
    MT_RECV_LINK_MSG = 11054
    """接收链接消息"""
    MT_RECV_FILE_MSG = 11055
    """接收文件消息"""
    MT_RECV_MINIAPP_MSG = 11056
    """接收小程序消息"""
    MT_RECV_WCPAY_MSG = 11057
    """接收好友转账消息"""
    MT_RECV_SYSTEM_MSG = 11058
    """接收系统消息"""
    MT_RECV_REVOKE_MSG = 11059
    """接收撤回消息"""
    MT_RECV_OTHER_MSG = 11060
    """接收其他未知消息"""
    MT_RECV_OTHER_APP_MSG = 11061
    """接收应用类型未知消息"""
    MT_ROOM_ADD_MEMBER_NOTIFY_MSG = 11098
    """群员新增通知"""
    MT_ROOM_DEL_MEMBER_NOTIFY_MSG = 11099
    """群员删除通知"""
    MT_ROOM_INTIVTED_NOTIFY_MSG = 11100
    """被邀请入群通知"""
    MT_FRIEND_ADD_NOTIFY_MSG = 11102
    """好友添加通知"""


class WxType(IntEnum):
    """微信原始类型枚举"""

    WX_MSG_TEXT = 1
    """文本"""
    WX_MSG_PICTURE = 3
    """图片"""
    WX_MSG_VOICE = 34
    """语音"""
    WX_MSG_FRIEND = 37
    """加好友请求"""
    WX_MSG_CARD = 42
    """名片"""
    WX_MSG_VIDEO = 43
    """视频"""
    WX_MSG_EMOJI = 47
    """表情"""
    WX_MSG_LOCATION = 48
    """位置"""
    WX_MSG_APP = 49
    """应用类型"""
    WX_MSG_SYSTEM = 10000
    """系统消息"""
    WX_MSG_REVOKE = 10002
    """撤回消息"""


class SubType(IntEnum):
    """应用子类型枚举"""

    WX_APPMSG_LINK = 5
    """链接（包含群邀请）"""
    WX_APPMSG_FILE = 6
    """文件"""
    WX_APPMSG_EMOJI = 8
    """表情消息"""
    WX_APPMSG_MUTIL = 19
    """合并消息"""
    WX_APPMSG_MINIAPP = 33
    """小程序"""
    WX_APPMSG_QUOTE = 57
    """引用消息"""
    WX_APPMSG_WCPAY = 2000
    """转账"""



"""ntchat适配器
适配ntchat服务
"""
import asyncio
import contextlib
import inspect
import json
from typing import Any, Dict, List, Optional, cast

from nonebot.drivers.fastapi import Driver
from nonebot.exception import ApiNotAvailable, NetworkError, WebSocketClosed
from nonebot.internal.driver import (
    URL,
    ForwardDriver,
    HTTPServerSetup,
    Request,
    Response,
    WebSocket,
    WebSocketServerSetup,
)
from nonebot.typing import overrides
from nonebot.utils import DataclassEncoder, escape_tag

from nonebot.adapters import Adapter as BaseAdapter

from . import event
from .bot import Bot
from .collator import EventModels
from .config import Config
from .event import Event
from .store import ResultStore
from .utils import handle_api_result, log

event_models = EventModels[Event]()
"""事件模型创建器"""


class Adapter(BaseAdapter):

    _result_store = ResultStore()
    """api回调存储"""
    ntchat_config: Config
    """ntchat配置"""

    def __init__(self, driver: Driver, **kwargs) -> None:
        super().__init__(driver, **kwargs)
        self.ntchat_config: Config = Config(**self.config.dict())
        self.connections: Dict[str, WebSocket] = {}
        self.tasks: List["asyncio.Task"] = []
        self._search_events()
        self._setup()

    def _search_events(self) -> None:
        """搜索事件模型"""
        for model_name in dir(event):
            model = getattr(event, model_name)
            if not inspect.isclass(model) or not issubclass(model, Event):
                continue
            event_models.add_event_model(model)

    def _setup(self) -> None:
        http_setup = HTTPServerSetup(
            URL("/ntchat/"), "POST", self.get_name(), self._handle_http
        )
        self.setup_http_server(http_setup)
        http_setup = HTTPServerSetup(
            URL("/ntchat/http"), "POST", self.get_name(), self._handle_http
        )
        self.setup_http_server(http_setup)
        http_setup = HTTPServerSetup(
            URL("/ntchat/http/"), "POST", self.get_name(), self._handle_http
        )
        self.setup_http_server(http_setup)

        ws_setup = WebSocketServerSetup(
            URL("/ntchat/"), self.get_name(), self._handle_ws
        )
        self.setup_websocket_server(ws_setup)
        ws_setup = WebSocketServerSetup(
            URL("/ntchat/ws"), self.get_name(), self._handle_ws
        )
        self.setup_websocket_server(ws_setup)
        ws_setup = WebSocketServerSetup(
            URL("/ntchat/ws/"), self.get_name(), self._handle_ws
        )
        self.setup_websocket_server(ws_setup)

    @classmethod
    @overrides(BaseAdapter)
    def get_name(cls) -> str:
        """适配器名称: `ntchat`"""
        return "ntchat"

    @overrides(BaseAdapter)
    async def _call_api(self, bot: Bot, api: str, **data: Any) -> Any:
        websocket = self.connections.get(bot.self_id, None)
        timeout: float = data.get("_timeout", self.config.api_timeout)
        log("DEBUG", f"Calling API <y>{api}</y>")

        if websocket:
            seq = self._result_store.get_seq()
            json_data = json.dumps(
                {"action": api, "params": data, "echo": str(seq)},
                cls=DataclassEncoder,
                ensure_ascii=False,
            )
            await websocket.send(json_data)
            return handle_api_result(
                await self._result_store.fetch(bot.self_id, seq, timeout)
            )
        elif isinstance(self.driver, ForwardDriver):
            api_root = self.ntchat_config.ntchat_http_api_root
            if not api_root:
                raise ApiNotAvailable
            elif not api_root.endswith("/"):
                api_root += "/"

            headers = {"Content-Type": "application/json"}

            request = Request(
                "POST",
                api_root + api,
                headers=headers,
                timeout=timeout,
                content=json.dumps(data, cls=DataclassEncoder),
            )

            try:
                response = await self.driver.request(request)

                if 200 <= response.status_code < 300:
                    if not response.content:
                        raise ValueError("Empty response")
                    result = json.loads(response.content)
                    return handle_api_result(result)
                raise NetworkError(
                    f"HTTP request received unexpected "
                    f"status code: {response.status_code}"
                )
            except NetworkError:
                raise
            except Exception as e:
                raise NetworkError("HTTP request failed") from e
        else:
            raise ApiNotAvailable

    async def _handle_http(self, request: Request) -> Response:
        self_id = request.headers.get("X-Self-ID")

        # check self_id
        if not self_id:
            log("WARNING", "Missing X-Self-ID Header")
            return Response(400, content="Missing X-Self-ID Header")

        # check access_token
        response = self._check_access_token(request)
        if response is not None:
            return response

        data = request.content
        if data is not None:
            json_data = json.loads(data)
            event = self.json_to_event(json_data)
            if event:
                bot = self.bots.get(self_id, None)
                if not bot:
                    bot = Bot(self, self_id)
                    self.bot_connect(bot)
                    log("INFO", f"<y>Bot {escape_tag(self_id)}</y> connected")
                bot = cast(Bot, bot)
                asyncio.create_task(bot.handle_event(event))
        return Response(204)

    async def _handle_ws(self, websocket: WebSocket) -> None:
        self_id = websocket.request.headers.get("X-Self-ID")

        # check self_id
        if not self_id:
            log("WARNING", "Missing X-Self-ID Header")
            await websocket.close(1008, "Missing X-Self-ID Header")
            return
        elif self_id in self.bots:
            log("WARNING", f"There's already a bot {self_id}, ignored")
            await websocket.close(1008, "Duplicate X-Self-ID")
            return

        # check access_token
        response = self._check_access_token(websocket.request)
        if response is not None:
            content = cast(str, response.content)
            await websocket.close(1008, content)
            return

        await websocket.accept()
        bot = Bot(self, self_id)
        self.connections[self_id] = websocket
        self.bot_connect(bot)

        log("INFO", f"<y>Bot {escape_tag(self_id)}</y> connected")

        try:
            while True:
                data = await websocket.receive()
                json_data = json.loads(data)
                event = self.json_to_event(json_data, self_id)
                if event:
                    asyncio.create_task(bot.handle_event(event))
        except WebSocketClosed:
            log("WARNING", f"WebSocket for Bot {escape_tag(self_id)} closed by peer")
        except Exception as e:
            log(
                "ERROR",
                f"<r><bg #f8bbd0>Error while process data from websocketfor bot {escape_tag(self_id)}.</bg #f8bbd0></r>",
                e,
            )
        finally:
            with contextlib.suppress(Exception):
                await websocket.close()
            self.connections.pop(self_id, None)
            self.bot_disconnect(bot)

    def _check_access_token(self, request: Request) -> Optional[Response]:
        token = request.headers.get("access_token")

        access_token = self.ntchat_config.access_token
        if access_token and access_token != token:
            msg = "身份认证失败" if token else "缺少身份认证码"
            log("WARNING", msg)
            return Response(403, content=msg)

    @classmethod
    def json_to_event(
        cls, json_data: Any, self_id: Optional[str] = None
    ) -> Optional[Event]:
        """将 json 数据转换为 Event 对象。

        如果为 API 调用返回数据且提供了 Event 对应 Bot，则将数据存入 ResultStore。

        参数:
            json_data: json 数据
            self_id: 当前 Event 对应的 Bot

        返回:
            Event 对象，如果解析失败或为 API 调用返回数据，则返回 None
        """
        if not isinstance(json_data, dict):
            return None

        # api回调设置结果
        if "type" not in json_data:
            if self_id is not None:
                cls._result_store.add_result(self_id, json_data)
            return

        # 实例化事件
        event_model = event_models.get_event_model(json_data)
        try:
            json_data.update(**json_data.get("data"))
            event = event_model.parse_obj(json_data)
            return event
        except Exception as e:
            log(
                "ERROR",
                "<r><bg #f8bbd0>Failed to parse event. "
                f"Raw: {escape_tag(str(json_data))}</bg #f8bbd0></r>",
                e,
            )



import asyncio
import re
from io import BytesIO
from pathlib import Path
from typing import Any, Callable, Union

from nonebot.message import handle_event
from nonebot.typing import overrides

from nonebot.adapters import Bot as BaseBot

from .event import Event, TextMessageEvent
from .exception import NotInteractableEventError
from .message import Message, MessageSegment
from .utils import log


def _check_at_me(bot: "Bot", event: TextMessageEvent) -> None:
    """检查消息开头或结尾是否存在 @机器人，去除并赋值 `event.to_me`。

    参数:
        bot: Bot 对象
        event: TextMessageEvent 对象
    """
    if not isinstance(event, TextMessageEvent):
        return

    if bot.self_id in event.at_user_list:
        event.to_me = True


def _check_nickname(bot: "Bot", event: TextMessageEvent) -> None:
    """检查消息开头是否存在昵称，去除并赋值 `event.to_me`。

    参数:
        bot: Bot 对象
        event: TextMessageEvent 对象
    """
    first_text = event.msg
    nicknames = set(filter(lambda n: n, bot.config.nickname))
    if nicknames:
        # check if the user is calling me with my nickname
        nickname_regex = "|".join(nicknames)
        m = re.search(rf"^({nickname_regex})([\s,，]*|$)", first_text, re.IGNORECASE)
        if m:
            nickname = m.group(1)
            log("DEBUG", f"User is calling me {nickname}")
            event.to_me = True
            loc = m.end()
            event.msg = first_text[loc:]


async def send(
    bot: "Bot",
    event: Event,
    message: Union[str, MessageSegment, Message],
    **params: Any,  # extra options passed to send_msg API
) -> Any:
    """默认回复消息处理函数。"""
    try:
        from_wxid: str = getattr(event, "from_wxid")
        room_wxid: str = getattr(event, "room_wxid")
    except AttributeError:
        raise NotInteractableEventError("该类型事件没有交互对象，无法发送消息！")
    wx_id = from_wxid if room_wxid == "" else room_wxid

    if isinstance(message, str) or isinstance(message, MessageSegment):
        message = Message(message)

    task = []
    for segment in message:
        api = f"send_{segment.type}"
        segment.data["to_wxid"] = wx_id
        task.append(bot.call_api(api, **segment.data))

    return await asyncio.gather(*task)


class Bot(BaseBot):
    """
    ntchat协议适配。
    """

    send_handler: Callable[["Bot", Event, Union[str, MessageSegment]], Any] = send

    async def handle_event(self, event: Event) -> None:
        """处理收到的事件。"""
        if isinstance(event, TextMessageEvent):
            _check_at_me(self, event)
            _check_nickname(self, event)

        await handle_event(self, event)

    @overrides(BaseBot)
    async def send(
        self,
        event: Event,
        message: Union[str, MessageSegment],
        **kwargs: Any,
    ) -> Any:
        """根据 `event` 向触发事件的主体回复消息。

        参数:
            event: Event 对象
            message: 要发送的消息
            kwargs: 其他参数

        返回:
            API 调用返回数据

        异常:
            NetworkError: 网络错误
            ActionFailed: API 调用失败
        """
        return await self.__class__.send_handler(self, event, message, **kwargs)

    async def send_image(
        self, to_wxid: str, file_path: Union[str, bytes, BytesIO, Path]
    ):
        """
        说明:
            发送图片

        参数:
            * `to_wxid`：接收方的wx_id，可以是好友id，也可以是room_id
            * `file`：图片内容，支持url，本地路径，bytes，BytesIO
        """
        data = MessageSegment.image(file_path=file_path).data
        data["to_wxid"] = to_wxid
        return await self.call_api("send_image", **data)

    async def send_file(
        self, *, to_wxid: str, file_path: Union[str, bytes, BytesIO, Path]
    ):
        """
        说明:
            发送文件

        参数:
            * `to_wxid`：接收人id
            * `file_path`：文件内容，支持url，本地路径，bytes，BytesIO
        """
        data = MessageSegment.file(file_path=file_path).data
        data["to_wxid"] = to_wxid
        return await self.call_api("send_file", **data)

    async def send_video(
        self, *, to_wxid: str, file_path: Union[str, bytes, BytesIO, Path]
    ):
        """
        说明:
            发送视频

        参数:
            * `to_wxid`：接收人id
            * `file_path`：视频内容，支持url，本地路径，bytes，BytesIO
        """
        data = MessageSegment.video(file_path=file_path).data
        data["to_wxid"] = to_wxid
        return await self.call_api("send_video", **data)

    async def send_gif(self, *, to_wxid: str, file: Union[str, bytes, BytesIO, Path]):
        """
        说明:
            发送gif图片

        参数:
            * `to_wxid`：接收人id
            * `file`：图片内容，支持url，本地路径，bytes，BytesIO
        """
        data = MessageSegment.gif(file=file).data
        data["to_wxid"] = to_wxid
        return await self.call_api("send_gif", **data)



from typing import Optional

from pydantic import AnyUrl, BaseModel, Field


class WSUrl(AnyUrl):
    """ws或wss url"""

    allow_schemes = {"ws", "wss"}


class Config(BaseModel):
    """ntchat 配置类"""

    access_token: Optional[str] = Field(default=None)
    """令牌口令"""
    ntchat_http_api_root: Optional[str] = Field(default=None)
    """http api请求地址"""

    class Config:
        extra = "ignore"
